#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .livewindow import LiveWindow
from .resource import Resource
from .sensorbase import SensorBase
from .timer import Timer

__all__ = ["AnalogInput"]

class AnalogInput(SensorBase):
    """Analog input

    Each analog channel is read from hardware as a 12-bit number representing
    0V to 5V.

    Connected to each analog channel is an averaging and oversampling engine.
    This engine accumulates the specified (by :func:`setAverageBits` and
    :func:`setOversampleBits`) number of samples before returning a new value.
    This is not a sliding window average. The only difference between the
    oversampled samples and the averaged samples is that the oversampled
    samples are simply accumulated effectively increasing the resolution,
    while the averaged samples are divided by the number of samples to retain
    the resolution, but get more stable values.
    """

    kAccumulatorSlot = 1
    kAccumulatorChannels = (0, 1)
    channels = Resource(SensorBase.kAnalogInputChannels)

    def __init__(self, channel):
        """Construct an analog channel.
        :param channel: The channel number to represent. 0-3 are on-board 4-7 are on the MXP port.
        """
        if not hal.checkAnalogInputChannel(channel):
            raise IndexError("Analog input channel %d cannot be allocated. Channel is not present." % channel)
        try:
            AnalogInput.channels.allocate(self, channel)
        except IndexError as e:
            raise IndexError("Analog input channel %d is already allocated" % channel) from e

        self.channel = channel
        self.accumulatorOffset = 0

        port = hal.getPort(channel)
        self.port = hal.initializeAnalogInputPort(port)

        LiveWindow.addSensorChannel("AnalogInput", channel, self)
        hal.HALReport(hal.HALUsageReporting.kResourceType_AnalogChannel,
                      channel)

    def free(self):
        if self.channel is None:
            return
        AnalogInput.channels.free(self.channel)
        self.channel = None
        self.accumulatorOffset = 0

    def getValue(self):
        """Get a sample straight from this channel. The sample is a 12-bit
        value representing the 0V to 5V range of the A/D converter. The units
        are in A/D converter codes. Use :func:`getVoltage` to get the analog
        value in calibrated units.

        :returns: A sample straight from this channel.
        """
        return hal.getAnalogValue(self.port)

    def getAverageValue(self):
        """Get a sample from the output of the oversample and average engine
        for this channel. The sample is 12-bit + the bits configured in
        :func:`setOversampleBits`. The value configured in
        :func:`setAverageBits` will cause this value to be averaged 2**bits
        number of samples. This is not a sliding window. The sample will not
        change until 2^(OversampleBits + AverageBits) samples have been
        acquired from this channel. Use :func:`getAverageVoltage` to get the
        analog value in calibrated units.

        :returns: A sample from the oversample and average engine for this
            channel.
        """
        return hal.getAnalogAverageValue(self.port)

    def getVoltage(self):
        """Get a scaled sample straight from this channel. The value is scaled
        to units of Volts using the calibrated scaling data from
        :func:`getLSBWeight` and :func:`getOffset`.

        :returns: A scaled sample straight from this channel.
        """
        return hal.getAnalogVoltage(self.port)

    def getAverageVoltage(self):
        """Get a scaled sample from the output of the oversample and average
        engine for this channel. The value is scaled to units of Volts using
        the calibrated scaling data from :func:`getLSBWeight` and
        :func:`getOffset`. Using oversampling will cause this value to be
        higher resolution, but it will update more slowly. Using averaging
        will cause this value to be more stable, but it will update more
        slowly.

        :returns: A scaled sample from the output of the oversample and average
            engine for this channel.
        """
        return hal.getAnalogAverageVoltage(self.port)

    def getLSBWeight(self):
        """Get the factory scaling least significant bit weight constant. The
        least significant bit weight constant for the channel that was
        calibrated in manufacturing and stored in an eeprom.

        Volts = ((LSB_Weight * 1e-9) * raw) - (Offset * 1e-9)

        :returns: Least significant bit weight.
        """
        return hal.getAnalogLSBWeight(self.port)

    def getOffset(self):
        """Get the factory scaling offset constant. The offset constant for the
        channel that was calibrated in manufacturing and stored in an eeprom.

        Volts = ((LSB_Weight * 1e-9) * raw) - (Offset * 1e-9)

        :returns: Offset constant.
        """
        return hal.getAnalogOffset(self.port)

    def getChannel(self):
        """Get the channel number.

        :returns: The channel number.
        """
        return self.channel

    def setAverageBits(self, bits):
        """Set the number of averaging bits. This sets the number of
        averaging bits.  The actual number of averaged samples is 2^bits.
        The averaging is done automatically in the FPGA.

        :param bits: The number of averaging bits.
        """
        hal.setAnalogAverageBits(self.port, bits)

    def getAverageBits(self):
        """Get the number of averaging bits. This gets the number of averaging
        bits from the FPGA. The actual number of averaged samples is 2^bits.
        The averaging is done automatically in the FPGA.

        :returns: The number of averaging bits.
        """
        return hal.getAnalogAverageBits(self.port)

    def setOversampleBits(self, bits):
        """Set the number of oversample bits. This sets the number of
        oversample bits. The actual number of oversampled values is 2^bits.
        The oversampling is done automatically in the FPGA.

        :param bits: The number of oversample bits.
        """
        hal.setAnalogOversampleBits(self.port, bits)

    def getOversampleBits(self):
        """Get the number of oversample bits. This gets the number of
        oversample bits from the FPGA. The actual number of oversampled values
        is 2^bits.  The oversampling is done automatically in the FPGA.

        :returns: The number of oversample bits.
        """
        return hal.getAnalogOversampleBits(self.port)

    def initAccumulator(self):
        """Initialize the accumulator.
        """
        if not self.isAccumulatorChannel():
            raise IndexError(
                    "Accumulators are only available on slot %d on channels %s"
                    % (AnalogInput.kAccumulatorSlot,
                       ",".join(str(c) for c in AnalogInput.kAccumulatorChannels)))
        self.accumulatorOffset = 0
        hal.initAccumulator(self.port)

    def setAccumulatorInitialValue(self, initialValue):
        """Set an initial value for the accumulator.

        This will be added to all values returned to the user.

        :param initialValue:
            The value that the accumulator should start from when reset.
        """
        self.accumulatorOffset = initialValue

    def resetAccumulator(self):
        """Resets the accumulator to the initial value.
        """
        hal.resetAccumulator(self.port)

        # Wait until the next sample, so the next call to getAccumulator*()
        # won't have old values.
        sampleTime = 1.0 / AnalogInput.getGlobalSampleRate()
        overSamples = 1 << self.getOversampleBits()
        averageSamples = 1 << self.getAverageBits()
        Timer.delay(sampleTime * overSamples * averageSamples)

    def setAccumulatorCenter(self, center):
        """Set the center value of the accumulator.

        The center value is subtracted from each A/D value before it is added
        to the accumulator. This is used for the center value of devices like
        gyros and accelerometers to make integration work and to take the
        device offset into account when integrating.

        This center value is based on the output of the oversampled and
        averaged source from channel 1. Because of this, any non-zero
        oversample bits will affect the size of the value for this field.
        """
        hal.setAccumulatorCenter(self.port, center)

    def setAccumulatorDeadband(self, deadband):
        """Set the accumulator's deadband.
        """
        hal.setAccumulatorDeadband(self.port, deadband)

    def getAccumulatorValue(self):
        """Read the accumulated value.

        Read the value that has been accumulating. The accumulator
        is attached after the oversample and average engine.

        :returns: The 64-bit value accumulated since the last :func:`reset`.
        """
        return hal.getAccumulatorValue(self.port) + self.accumulatorOffset

    def getAccumulatorCount(self):
        """Read the number of accumulated values.

        Read the count of the accumulated values since the accumulator was
        last :func:`reset`.

        :returns: The number of times samples from the channel were
            accumulated.
        """
        return hal.getAccumulatorCount(self.port)

    def getAccumulatorOutput(self):
        """Read the accumulated value and the number of accumulated values
        atomically.

        This function reads the value and count from the FPGA atomically. This
        can be used for averaging.

        :returns: tuple of (value, count)
        """
        if not self.isAccumulatorChannel():
            raise IndexError("Channel %d is not an accumulator channel." % self.channel)
        return hal.getAccumulatorOutput(self.port)

    def isAccumulatorChannel(self):
        """Is the channel attached to an accumulator.

        :returns: The analog channel is attached to an accumulator.
        """
        return self.channel in AnalogInput.kAccumulatorChannels

    @staticmethod
    def setGlobalSampleRate(samplesPerSecond):
        """Set the sample rate per channel.

        This is a global setting for all channels.
        The maximum rate is 500kS/s divided by the number of channels in use.
        This is 62500 samples/s per channel if all 8 channels are used.

        :param samplesPerSecond: The number of samples per second.
        """
        hal.setAnalogSampleRate(float(samplesPerSecond))

    @staticmethod
    def getGlobalSampleRate():
        """Get the current sample rate.

        This assumes one entry in the scan list. This is a global setting for
        all channels.

        :returns: Sample rate.
        """
        return hal.getAnalogSampleRate()

    def pidGet(self):
        """Get the average voltage for use with PIDController

        :returns: the average voltage
        """
        return self.getAverageVoltage()

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Analog Input"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.getAverageVoltage())

    def startLiveWindowMode(self):
        # Analog Channels don't have to do anything special when entering the
        # LiveWindow.
        pass

    def stopLiveWindowMode(self):
        # Analog Channels don't have to do anything special when exiting the
        # LiveWindow.
        pass



import hal

from .sensorbase import SensorBase

__all__ = ["Compressor"]

class Compressor(SensorBase):
    """Operates the PCM (Pneumatics compressor module)
    
    The PCM automatically will run in close-loop mode by default whenever a
    Solenoid object is created. For most cases the Compressor object does not
    need to be instantiated or used in a robot program.

    This class is only required in cases where more detailed status or to
    enable/disable closed loop control. Note: you cannot operate the
    compressor directly from this class as doing so would circumvent the
    safety provided in using the pressure switch and closed loop control.
    You can only turn off closed loop control, thereby stopping the
    compressor from operating.
    
    .. not_implemented: initCompressor
    """

    def __init__(self, pcmId=None):
        """Create an instance of the Compressor
        
        :param pcmID: The PCM CAN device ID. Most robots that use PCM will
                      have a single module. Use this for supporting a second
                      module other than the default.
        """
        if pcmId is None:
            pcmId = SensorBase.getDefaultSolenoidModule()
        self.pcm = hal.initializeCompressor(pcmId)

    def start(self):
        """Start the compressor running in closed loop control mode.
        Use the method in cases where you would like to manually stop and
        start the compressor for applications such as conserving battery
        or making sure that the compressor motor doesn't start during
        critical operations.
        """
        self.setClosedLoopControl(True)

    def stop(self):
        """Stop the compressor from running in closed loop control mode.
        Use the method in cases where you would like to manually stop and
        start the compressor for applications such as conserving battery
        or making sure that the compressor motor doesn't start during
        critical operations.
        """
        self.setClosedLoopControl(False)

    def enabled(self):
        """Get the enabled status of the compressor.
        
        :returns: True if the compressor is on
        :rtype: bool
        """
        return hal.getCompressor(self.pcm)

    def getPressureSwitchValue(self):
        """Get the current pressure switch value.
        
        :returns: True if the pressure is low by reading the pressure switch
            that is plugged into the PCM
        :rtype: bool
        """
        return hal.getPressureSwitch(self.pcm)

    def getCompressorCurrent(self):
        """Get the current being used by the compressor.
        
        :returns: Current consumed in amps for the compressor motor
        :rtype: float
        """
        return hal.getCompressorCurrent(self.pcm)

    def setClosedLoopControl(self, on):
        """Set the PCM in closed loop control mode.
        
        :param on: If True sets the compressor to be in closed loop control
                   mode otherwise normal operation of the compressor is disabled.
        :type  on: bool
        """
        hal.setClosedLoopControl(self.pcm, on)

    def getClosedLoopControl(self):
        """Gets the current operating mode of the PCM.
        
        :returns: True if compressor is operating on closed-loop mode,
                  otherwise return False.
        :rtype: bool
        """
        return hal.getClosedLoopControl(self.pcm)

    def getCompressorCurrentTooHighFault(self):
        """
        :returns: True if PCM is in fault state : Compressor Drive is
            disabled due to compressor current being too high
        """
        return hal.getCompressorCurrentTooHighFault(self.pcm)

    def getCompressorCurrentTooHighStickyFault(self):
        """
        :returns: True if PCM sticky fault is set : Compressor Drive is
            disabled due to compressor current being too high
        """
        return hal.getCompressorCurrentTooHighStickyFault(self.pcm)

    def getCompressorShortedFault(self):
        """
        :returns: True if PCM is in fault state : Compressor Output
            appears to be shorted
        """
        return hal.getCompressorShortedFault(self.pcm)

    def getCompressorShortedStickyFault(self):
        """
        :returns: True if PCM sticky fault is set : Compressor Output
            appears to be shorted
        """
        return hal.getCompressorShortedStickyFault(self.pcm)

    def getCompressorNotConnectedFault(self):
        """
        :returns: True if PCM is in fault state : Compressor does not appear
            to be wired, i.e. compressor is not drawing enough current.
        """
        return hal.getCompressorNotConnectedFault(self.pcm)

    def getCompressorNotConnectedStickyFault(self):
        """
        :returns: True if PCM sticky fault is set : Compressor does not appear
            to be wired, i.e. compressor is not drawing enough current.
        """
        return hal.getCompressorNotConnectedStickyFault(self.pcm)

    def clearAllPCMStickyFaults(self):
        hal.clearAllPCMStickyFaults(self.pcm)

    def getSmartDashboardType(self):
        return "Compressor"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putBoolean("Enabled", self.enabled())
            table.putBoolean("Pressure Switch", self.getPressureSwitchValue())



import hal
import weakref

__all__ = ["SPI"]

def _freeSPI(port):
    hal.spiClose(port)

class SPI:
    """Represents a SPI bus port"""

    class Port:
        kOnboardCS0 = 0
        kOnboardCS1 = 1
        kOnboardCS2 = 2
        kOnboardCS3 = 3
        kMXP = 4

    devices = 0
    
    @staticmethod
    def _reset():
        SPI.devices = 0

    def __init__(self, port):
        """Constructor

        :param port: the physical SPI port
        """
        self.port = port
        self.bitOrder = 0
        self.clockPolarity = 0
        self.dataOnTrailing = 0

        hal.spiInitialize(self.port)
        self._spi_finalizer = weakref.finalize(self, _freeSPI, self.port)

        SPI.devices += 1
        hal.HALReport(hal.HALUsageReporting.kResourceType_SPI, SPI.devices)

    def setClockRate(self, hz):
        """Configure the rate of the generated clock signal.
        The default value is 500,000 Hz.
        The maximum value is 4,000,000 Hz.

        :param hz: The clock rate in Hertz.
        """
        hal.spiSetSpeed(self.port, hz)

    def setMSBFirst(self):
        """Configure the order that bits are sent and received on the wire
        to be most significant bit first.
        """
        self.bitOrder = 1
        hal.spiSetOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setLSBFirst(self):
        """Configure the order that bits are sent and received on the wire
        to be least significant bit first.
        """
        self.bitOrder = 0
        hal.spiSetOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setClockActiveLow(self):
        """Configure the clock output line to be active low.
        This is sometimes called clock polarity high or clock idle high.
        """
        self.clockPolarity = 1
        hal.spiSetOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setClockActiveHigh(self):
        """Configure the clock output line to be active high.
        This is sometimes called clock polarity low or clock idle low.
        """
        self.clockPolarity = 0
        hal.spiSetOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setSampleDataOnFalling(self):
        """Configure that the data is stable on the falling edge and the data
        changes on the rising edge.
        """
        self.dataOnTrailing = 1
        hal.spiSetOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setSampleDataOnRising(self):
        """Configure that the data is stable on the rising edge and the data
        changes on the falling edge.
        """
        self.dataOnTrailing = 0
        hal.spiSetOpts(self.port, self.bitOrder, self.dataOnTrailing,
                       self.clockPolarity)

    def setChipSelectActiveHigh(self):
        """Configure the chip select line to be active high.
        """
        hal.spiSetChipSelectActiveHigh(self.port)

    def setChipSelectActiveLow(self):
        """Configure the chip select line to be active low.
        """
        hal.spiSetChipSelectActiveLow(self.port)

    def write(self, dataToSend):
        """Write data to the slave device.  Blocks until there is space in the
        output FIFO.

        If not running in output only mode, also saves the data received
        on the MISO input during the transfer into the receive FIFO.

        :param dataToSend: Data to send (bytes)
        """
        hal.spiWrite(self.port, dataToSend)

    def read(self, initiate, size):
        """Read a word from the receive FIFO.

        Waits for the current transfer to complete if the receive FIFO is
        empty.

        If the receive FIFO is empty, there is no active transfer, and
        initiate is False, errors.

        :param initiate: If True, this function pushes "0" into the
            transmit buffer and initiates a transfer.  If False, this function
            assumes that data is already in the receive FIFO from a previous
            write.
        :param size: Number of bytes to read.

        :returns: received data bytes
        """
        if initiate:
            return hal.spiTransaction(self.port, [0]*size)
        else:
            return hal.spiRead(self.port, size)

    def transaction(self, dataToSend):
        """Perform a simultaneous read/write transaction with the device

        :param dataToSend: The data to be written out to the device

        :returns: data received from the device
        """
        return hal.spiTransaction(self.port, dataToSend)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

__all__ = ["SmartDashboard"]

class SmartDashboard:
    """The bridge between robot programs and the SmartDashboard on the laptop

    When a value is put into the SmartDashboard, it pops up on the
    SmartDashboard on the remote host. Users can put values into and get values
    from the SmartDashboard.
    
    These values can also be accessed by a NetworkTables client via the
    'SmartDashboard' table::
    
        from networktables import NetworkTable
        sd = NetworkTable.getTable('SmartDashboard')
        
        # sd.putXXX and sd.getXXX work as expected here
    
    """
    # The NetworkTable used by SmartDashboard
    table = None
    # A table linking tables in the SmartDashboard to the SmartDashboardData
    # objects they came from.
    tablesToData = {}

    class _defaultValueSentry:
        pass
    
    @staticmethod
    def _reset():
        SmartDashboard.tablesToData = {}
        SmartDashboard.table = None

    @staticmethod
    def getTable():
        if SmartDashboard.table is None:
            from networktables import NetworkTable
            SmartDashboard.table = NetworkTable.getTable("SmartDashboard")
            hal.HALReport(hal.HALUsageReporting.kResourceType_SmartDashboard,
                          hal.HALUsageReporting.kSmartDashboard_Instance)
        return SmartDashboard.table

    @staticmethod
    def putData(*args, **kwargs):
        """Maps the specified key to the specified value in this table.
        The value can be retrieved by calling the get method with a key that
        is equal to the original key.

        Two argument formats are supported: key, data:
        
        :param key: the key (cannot be None)
        :type  key: str
        :param data: the value

        Or the single argument "value":
        
        :param value: the named value (getName is called to retrieve the value)
        """
        # NOTE: mix of args and kwargs not allowed
        if kwargs and not args:
            if "value" in kwargs:
                data = kwargs["value"]
                key = data.getName()
            else:
                key = kwargs["key"]
                data = kwargs["data"]
        elif len(args) == 1 and not kwargs:
            data = args[0]
            key = data.getName()
        elif len(args) == 2 and not kwargs:
            key, data = args
        else:
            raise ValueError("only key, data or value accepted")

        table = SmartDashboard.getTable()
        dataTable = table.getSubTable(key)
        dataTable.putString("~TYPE~", data.getSmartDashboardType())
        data.initTable(dataTable)
        SmartDashboard.tablesToData[dataTable] = data

    @staticmethod
    def getData(key):
        """Returns the value at the specified key.
        
        :param key: the key
        :type  key: str
        :returns: the value
        
        :raises: :exc:`KeyError` if the key doesn't exist
        """
        table = SmartDashboard.getTable()
        subtable = table.getSubTable(key)
        data = SmartDashboard.tablesToData.get(subtable)
        if data is None:
            raise KeyError("SmartDashboard data does not exist: '%s'" % key)
        return data

    @staticmethod
    def putBoolean(key, value):
        """Maps the specified key to the specified value in this table.
        The key can not be None.
        
        The value can be retrieved by calling the get method with a key that
        is equal to the original key.
        
        :param key: the key
        :type  key: str
        :param value: the value
        """
        table = SmartDashboard.getTable()
        table.putBoolean(key, value)

    @staticmethod
    def getBoolean(key, defaultValue=_defaultValueSentry):
        """Returns the value at the specified key.
        
        :param key: the key
        :type  key: str
        :param defaultValue: returned if the key doesn't exist
        :returns: the value
        
        :raises: :exc:`KeyError` if the key doesn't exist and defaultValue
                 is not provided.
        """
        table = SmartDashboard.getTable()
        if defaultValue is SmartDashboard._defaultValueSentry:
            return table.getBoolean(key)
        else:
            return table.getBoolean(key, defaultValue)

    @staticmethod
    def putNumber(key, value):
        """Maps the specified key to the specified value in this table.
        The key can not be None.
        The value can be retrieved by calling the get method with a key that
        is equal to the original key.
        
        :param key: the key
        :type  key: str
        :param value: the value
        :type  value: int or float
        """
        table = SmartDashboard.getTable()
        table.putNumber(key, value)

    @staticmethod
    def getNumber(key, defaultValue=_defaultValueSentry):
        """Returns the value at the specified key.
        
        :param key: the key
        :type  key: str
        :param defaultValue: returned if the key doesn't exist
        :rtype: float
        
        :raises: :exc:`KeyError` if the key doesn't exist and defaultValue
                 is not provided.
        """
        table = SmartDashboard.getTable()
        if defaultValue is SmartDashboard._defaultValueSentry:
            return table.getNumber(key)
        else:
            return table.getNumber(key, defaultValue)

    @staticmethod
    def putString(key, value):
        """Maps the specified key to the specified value in this table.
        The key can not be None.
        The value can be retrieved by calling the get method with a key that
        is equal to the original key.
        
        :param key: the key
        :type  key: str
        :param value: the value
        :type  value: str
        """
        table = SmartDashboard.getTable()
        table.putString(key, value)

    @staticmethod
    def getString(key, defaultValue=_defaultValueSentry):
        """Returns the value at the specified key.
        
        :param key: the key
        :type  key: str
        :param defaultValue: returned if the key doesn't exist
        :rtype: str
        
        :raises: :exc:`KeyError` if the key doesn't exist and defaultValue
                 is not provided.
        """
        table = SmartDashboard.getTable()
        if defaultValue is SmartDashboard._defaultValueSentry:
            return table.getString(key)
        else:
            return table.getString(key, defaultValue)

    # Deprecated Methods
    putInt = putNumber
    getInt = getNumber
    putDouble = putNumber
    getDouble = getNumber



'''

    The WPI Robotics library (WPILib) is a set of classes that interfaces to the hardware in the FRC
    control system and your robot. There are classes to handle sensors, motors, the driver
    station, and a number of other utility functions like timing and field management.
    The library is designed to:
    
    - Deal with all the low level interfacing to these components so you can concentrate on
      solving this year's "robot problem". This is a philosophical decision to let you focus
      on the higher-level design of your robot rather than deal with the details of the
      processor and the operating system.
    - Understand everything at all levels by making the full source code of the library
      available. You can study (and modify) the algorithms used by the gyro class for
      oversampling and integration of the input signal or just ask the class for the current
      robot heading. You can work at any level.

'''

from .adxl345_i2c import *
from .adxl345_spi import *
from .analogaccelerometer import *
from .analoginput import *
from .analogoutput import *
from .analogpotentiometer import *
from .analogtrigger import *
from .analogtriggeroutput import *
from .builtinaccelerometer import *
from .canjaguar import *
from .cantalon import *
from .compressor import *
from .controllerpower import *
from .counter import *
from .digitalinput import *
from .digitaloutput import *
from .digitalsource import *
from .doublesolenoid import *
from .driverstation import *
from .encoder import *
from .geartooth import *
from .gyro import *
from .i2c import *
from .interruptablesensorbase import *
from .iterativerobot import *
from .jaguar import *
from .joystick import *
from .livewindow import *
from .livewindowsendable import *
from .motorsafety import *
from .pidcontroller import *
from .powerdistributionpanel import *
from .preferences import *
from .pwm import *
from .relay import *
from .resource import *
from .robotbase import *
from .robotdrive import *
from .robotstate import *
from .safepwm import *
from .samplerobot import *
from .sendable import *
from .sendablechooser import *
from .sensorbase import *
from .servo import *
from .smartdashboard import *
from .solenoidbase import *
from .solenoid import *
from .spi import *
from .talon import *
from .talonsrx import *
from .timer import *
from .ultrasonic import *
from .utility import *
from .victor import *
from .victorsp import *

# Provide dummy implementations if pynivision isn't available.
try:
    from .cameraserver import *
    from .usbcamera import *
except ImportError:
    from ._impl.dummycamera import *

from ._impl.main import run

try:
    from .version import __version__
except ImportError:
    __version__ = 'master'



# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.

import hal
import logging

from .robotbase import RobotBase
from .timer import Timer
from .livewindow import LiveWindow

__all__ = ["IterativeRobot"]

class IterativeRobot(RobotBase):
    """IterativeRobot implements a specific type of Robot Program framework,
    extending the :class:`.RobotBase` class.

    The IterativeRobot class is intended to be subclassed by a user creating a
    robot program.

    This class is intended to implement the "old style" default code, by
    providing the following functions which are called by the main loop,
    :meth:`startCompetition`, at the appropriate times:

    - :meth:`robotInit` -- provide for initialization at robot power-on

    init() functions -- each of the following functions is called once when the
    appropriate mode is entered:

    - :meth:`disabledInit`   -- called only when first disabled
    - :meth:`autonomousInit` -- called each and every time autonomous is entered from another mode
    - :meth:`teleopInit`     -- called each and every time teleop is entered from another mode
    - :meth:`testInit`       -- called each and every time test mode is entered from another mode

    Periodic() functions -- each of these functions is called iteratively at
    the appropriate periodic rate (aka the "slow loop").  The period of
    the iterative robot is synced to the driver station control packets,
    giving a periodic frequency of about 50Hz (50 times per second).
    
    - :meth:`disabledPeriodic`
    - :meth:`autonomousPeriodic`
    - :meth:`teleopPeriodic`
    - :meth:`testPeriodic`
    
    """
     
    #: A python logging object that you can use to send messages to the log. It
    #: is recommended to use this instead of print statements.
    logger = logging.getLogger("robot")

    def __init__(self):
        """Constructor for RobotIterativeBase.

        The constructor initializes the instance variables for the robot to
        indicate the status of initialization for disabled, autonomous, and
        teleop code.
        """
        super().__init__()
        # set status for initialization of disabled, autonomous, and teleop code.
        self.disabledInitialized = False
        self.autonomousInitialized = False
        self.teleopInitialized = False
        self.testInitialized = False

    def prestart(self):
        """Don't immediately say that the robot's ready to be enabled, see below"""
        pass

    def startCompetition(self):
        """Provide an alternate "main loop" via startCompetition()."""
        hal.HALReport(hal.HALUsageReporting.kResourceType_Framework,
                      hal.HALUsageReporting.kFramework_Iterative)

        self.robotInit()

        #We call this now (not in prestart like default) so that the robot
        #won't enable until the initialization has finished. This is useful
        #because otherwise it's sometimes possible to enable the robot before
        #the code is ready.
        hal.HALNetworkCommunicationObserveUserProgramStarting()

        # loop forever, calling the appropriate mode-dependent function
        LiveWindow.setEnabled(False)
        while True:
            # Call the appropriate function depending upon the current robot mode
            if self.isDisabled():
                # call DisabledInit() if we are now just entering disabled mode from
                # either a different mode or from power-on
                if not self.disabledInitialized:
                    LiveWindow.setEnabled(False)
                    self.disabledInit()
                    self.disabledInitialized = True
                    # reset the initialization flags for the other modes
                    self.autonomousInitialized = False
                    self.teleopInitialized = False
                    self.testInitialized = False
                if self.nextPeriodReady():
                    hal.HALNetworkCommunicationObserveUserProgramDisabled()
                    self.disabledPeriodic()
            elif self.isTest():
                # call TestInit() if we are now just entering test mode from either
                # a different mode or from power-on
                if not self.testInitialized:
                    LiveWindow.setEnabled(True)
                    self.testInit()
                    self.testInitialized = True
                    self.autonomousInitialized = False
                    self.teleopInitialized = False
                    self.disabledInitialized = False
                if self.nextPeriodReady():
                    hal.HALNetworkCommunicationObserveUserProgramTest()
                    self.testPeriodic()
            elif self.isAutonomous():
                # call Autonomous_Init() if this is the first time
                # we've entered autonomous_mode
                if not self.autonomousInitialized:
                    LiveWindow.setEnabled(False)
                    # KBS NOTE: old code reset all PWMs and relays to "safe values"
                    # whenever entering autonomous mode, before calling
                    # "Autonomous_Init()"
                    self.autonomousInit()
                    self.autonomousInitialized = True
                    self.testInitialized = False
                    self.teleopInitialized = False
                    self.disabledInitialized = False
                if self.nextPeriodReady():
                    hal.HALNetworkCommunicationObserveUserProgramAutonomous()
                    self.autonomousPeriodic()
            else:
                # call Teleop_Init() if this is the first time
                # we've entered teleop_mode
                if not self.teleopInitialized:
                    LiveWindow.setEnabled(False)
                    self.teleopInit()
                    self.teleopInitialized = True
                    self.testInitialized = False
                    self.autonomousInitialized = False
                    self.disabledInitialized = False
                if self.nextPeriodReady():
                    hal.HALNetworkCommunicationObserveUserProgramTeleop()
                    self.teleopPeriodic()
            self.ds.waitForData()

    def nextPeriodReady(self):
        """Determine if the appropriate next periodic function should be
        called.  Call the periodic functions whenever a packet is received
        from the Driver Station, or about every 20ms.
        
        :rtype: bool
        """
        return self.ds.isNewControlData()

    # ----------- Overridable initialization code -----------------

    def robotInit(self):
        """Robot-wide initialization code should go here.

        Users should override this method for default Robot-wide initialization
        which will be called when the robot is first powered on.  It will be
        called exactly 1 time.
        
        .. note:: It is simpler to override this function instead of defining
                  a constructor for your robot class
        """
        self.logger.info("Default IterativeRobot.robotInit() method... Overload me!")

    def disabledInit(self):
        """Initialization code for disabled mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters disabled mode.
        """
        self.logger.info("Default IterativeRobot.disabledInit() method... Overload me!")

    def autonomousInit(self):
        """Initialization code for autonomous mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters autonomous mode.
        """
        self.logger.info("Default IterativeRobot.autonomousInit() method... Overload me!")

    def teleopInit(self):
        """Initialization code for teleop mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters teleop mode.
        """
        self.logger.info("Default IterativeRobot.teleopInit() method... Overload me!")

    def testInit(self):
        """Initialization code for test mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters test mode.
        """
        self.logger.info("Default IterativeRobot.testInit() method... Overload me!")

    # ----------- Overridable periodic code -----------------

    def disabledPeriodic(self):
        """Periodic code for disabled mode should go here.

        Users should override this method for code which will be called
        periodically at a regular rate while the robot is in disabled mode.
        """
        func = self.disabledPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.info("Default IterativeRobot.disabledPeriodic() method... Overload me!")
            func.firstRun = False
        Timer.delay(0.001)

    def autonomousPeriodic(self):
        """Periodic code for autonomous mode should go here.

        Users should override this method for code which will be called
        periodically at a regular rate while the robot is in autonomous mode.
        """
        func = self.autonomousPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.info("Default IterativeRobot.autonomousPeriodic() method... Overload me!")
            func.firstRun = False
        Timer.delay(0.001)

    def teleopPeriodic(self):
        """Periodic code for teleop mode should go here.

        Users should override this method for code which will be called
        periodically at a regular rate while the robot is in teleop mode.
        """
        func = self.teleopPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.warn("Default IterativeRobot.teleopPeriodic() method... Overload me!")
            func.firstRun = False
        Timer.delay(0.001)

    def testPeriodic(self):
        """Periodic code for test mode should go here.

        Users should override this method for code which will be called
        periodically at a regular rate while the robot is in test mode.
        """
        func = self.testPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.info("Default IterativeRobot.testPeriodic() method... Overload me!")
            func.firstRun = False



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .command import Command

__all__ = ["CommandGroup"]

class CommandGroup(Command):
    """A CommandGroup is a list of commands which are executed in sequence.

    Commands in a CommandGroup are added using the :meth:`addSequential` method
    and are called sequentially. CommandGroups are themselves Commands and can
    be given to other CommandGroups.

    CommandGroups will carry all of the requirements of their subcommands.
    Additional requirements can be specified by calling :meth:`requires`
    normally in the constructor.

    CommandGroups can also execute commands in parallel, simply by adding them
    using addParallel(...).

    .. seealso:: :class:`.Command`, :class:`Subsystem`
    """

    class Entry:
        IN_SEQUENCE = 0
        BRANCH_PEER = 1
        BRANCH_CHILD = 2

        def __init__(self, command, state, timeout):
            self.command = command
            self.state = state
            self.timeout = timeout

        def isTimedOut(self):
            if self.timeout is None:
                return False
            else:
                time = self.command.timeSinceInitialized()
                if time == 0:
                    return False
                else:
                    return time >= self.timeout

    def __init__(self, name=None):
        """Creates a new CommandGroup with the given name.
        
        :param name: the name for this command group (optional).  If None,
                     the name of this command will be set to its class name.
        """
        super().__init__(name)
        # The commands in this group (stored in entries)
        self.commands = []
        # The active children in this group (stored in entries)
        self.children = []
        # The current command, None signifies that none have been run
        self.currentCommandIndex = None

    def addSequential(self, command, timeout=None):
        """Adds a new Command to the group (with an optional timeout).
        The Command will be started after all the previously added Commands.

        Once the Command is started, it will be run until it finishes or the
        time expires, whichever is sooner (if a timeout is provided).  Note
        that the given Command will have no knowledge that it is on a timer.

        Note that any requirements the given Command has will be added to the
        group.  For this reason, a Command's requirements can not be changed
        after being added to a group.

        It is recommended that this method be called in the constructor.

        :param command: The Command to be added
        :param timeout: The timeout (in seconds) (optional)
        """
        with self.mutex:
            if self.locked:
                raise ValueError("Can not add new command to command group")
            if command is None:
                raise ValueError("Given None command")
            if timeout is not None and timeout < 0:
                raise ValueError("Can not be given a negative timeout")

            command.setParent(self)

            self.commands.append(CommandGroup.Entry(command, CommandGroup.Entry.IN_SEQUENCE, timeout))
            for reqt in command.getRequirements():
                self.requires(reqt)

    def addParallel(self, command, timeout=None):
        """Adds a new child Command to the group (with an optional timeout).
        The Command will be started after all the previously added Commands.

        Once the Command is started, it will run until it finishes, is
        interrupted, or the time expires (if a timeout is provided), whichever
        is sooner.  Note that the given Command will have no knowledge that it
        is on a timer.

        Instead of waiting for the child to finish, a CommandGroup will have it
        run at the same time as the subsequent Commands.  The child will run
        until either it finishes, the timeout expires, a new child with
        conflicting requirements is started, or the main sequence runs a
        Command with conflicting requirements.  In the latter two cases, the
        child will be canceled even if it says it can't be interrupted.

        Note that any requirements the given Command has will be added to the
        group.  For this reason, a Command's requirements can not be changed
        after being added to a group.

        It is recommended that this method be called in the constructor.

        :param command: The command to be added
        :param timeout: The timeout (in seconds) (optional)
        """
        with self.mutex:
            if self.locked:
                raise ValueError("Can not add new command to command group")
            if command is None:
                raise ValueError("Given null command")
            if timeout is not None and timeout < 0:
                raise ValueError("Can not be given a negative timeout")

            command.setParent(self)

            self.commands.append(CommandGroup.Entry(command, CommandGroup.Entry.BRANCH_CHILD, timeout))
            for reqt in command.getRequirements():
                self.requires(reqt)

    def _initialize(self):
        self.currentCommandIndex = None

    def _execute(self):
        entry = None
        cmd = None
        firstRun = False
        if self.currentCommandIndex is None:
            firstRun = True
            self.currentCommandIndex = 0

        while self.currentCommandIndex < len(self.commands):
            if cmd is not None:
                if entry.isTimedOut():
                    cmd._cancel()
                if cmd.run():
                    break
                else:
                    cmd.removed()
                    self.currentCommandIndex += 1
                    firstRun = True
                    cmd = None
                    continue

            entry = self.commands[self.currentCommandIndex]
            cmd = None

            if entry.state == CommandGroup.Entry.IN_SEQUENCE:
                cmd = entry.command
                if firstRun:
                    cmd.startRunning()
                    self.cancelConflicts(cmd)
                firstRun = False
            elif entry.state == CommandGroup.Entry.BRANCH_PEER:
                self.currentCommandIndex += 1
                entry.command.start()
            elif entry.state == CommandGroup.Entry.BRANCH_CHILD:
                self.currentCommandIndex += 1
                self.cancelConflicts(entry.command)
                entry.command.startRunning()
                self.children.append(entry)

        # Run Children
        toremove = []
        for i, entry in enumerate(self.children):
            child = entry.command
            if entry.isTimedOut():
                child._cancel()
            if not child.run():
                child.removed()
                toremove.append(i)
        for i in reversed(toremove):
            del self.children[i]

    def _end(self):
        # Theoretically, we don't have to check this, but we do if teams
        # override the isFinished method
        if self.currentCommandIndex is not None and self.currentCommandIndex < len(self.commands):
            cmd = self.commands[self.currentCommandIndex].command
            cmd._cancel()
            cmd.removed()

        for entry in self.children:
            cmd = entry.command
            cmd._cancel()
            cmd.removed()
        self.children.clear()

    def _interrupted(self):
        self._end()

    def isFinished(self):
        """Returns True if all the Commands in this group
        have been started and have finished.

        Teams may override this method, although they should probably
        reference super().isFinished() if they do.

        :returns: whether this CommandGroup is finished
        """
        if self.currentCommandIndex is None:
            return False
        return (self.currentCommandIndex >= len(self.commands) and
                not self.children)

    def initialize(self):
        pass # Can be overwritten by teams

    def execute(self):
        pass # Can be overwritten by teams

    def end(self):
        pass # Can be overwritten by teams

    def interrupted(self):
        pass # Can be overwritten by teams

    def isInterruptible(self):
        """Returns whether or not this group is interruptible.
        A command group will be uninterruptible if setInterruptable(False)
        was called or if it is currently running an uninterruptible command
        or child.

        :returns: whether or not this CommandGroup is interruptible.
        """
        with self.mutex:
            if not super().isInterruptible():
                return False

            if (self.currentCommandIndex is not None and
                self.currentCommandIndex < len(self.commands)):
                cmd = self.commands[self.currentCommandIndex].command
                if not cmd.isInterruptible():
                    return False

            for entry in self.children:
                if not entry.command.isInterruptible():
                    return False

        return True

    def cancelConflicts(self, command):
        toremove = []
        for i, entry in enumerate(self.children):
            child = entry.command

            for requirement in command.getRequirements():
                if child.doesRequire(requirement):
                    child._cancel()
                    child.removed()
                    toremove.append(i)
                    break

        for i in reversed(toremove):
            del self.children[i]



import nivision
import ctypes
import logging
import socket
import struct
import threading
import time

__all__ = ["CameraServer"]

logger = logging.getLogger(__name__)

class CameraServer:
    kPort = 1180
    kMagicNumber = bytes([0x01, 0x00, 0x00, 0x00])
    kSize640x480 = 0
    kSize320x240 = 1
    kSize160x120 = 2
    kHardwareCompression = -1
    kMaxImageSize = 200000

    intStruct = struct.Struct("!i")

    server = None

    @staticmethod
    def _reset():
        CameraServer.server = None

    @staticmethod
    def getInstance():
        if CameraServer.server is None:
            CameraServer.server = CameraServer()
        return CameraServer.server

    def __init__(self):
        self.mutex = threading.RLock()
        self.ready = threading.Event()
        self.dataPool = [(ctypes.c_ubyte * self.kMaxImageSize)(),
                         (ctypes.c_ubyte * self.kMaxImageSize)(),
                         (ctypes.c_ubyte * self.kMaxImageSize)()]
        self.imageData = (None, 0, 0, None)
        self.captureThread = None

        self.quality = 50
        self.hwClient = True
        self.camera = None

        self.serverThread = threading.Thread(target=self._serve, name="CameraServer")
        self.serverThread.daemon = True
        self.serverThread.start()

    def _freeImageData(self, imageData):
        if imageData[3] is not None:
            nivision.imaqDispose(imageData[3])
        elif imageData[0] is not None:
            with self.mutex:
                self.dataPool.append(imageData[0])

    def _setImageData(self, data, size, start=0, vdata=None):
        with self.mutex:
            self._freeImageData(self.imageData)
            self.imageData = (data, size, start, vdata)
        self.ready.set()

    def setImage(self, image):
        """Manually change the image that is served by the MJPEG stream. This
        can be called to pass custom annotated images to the dashboard. Note
        that, for 640x480 video, this method could take between 40 and 50
        milliseconds to complete.

        This shouldn't be called if startAutomaticCapture() is called.

        :param image: The IMAQ image to show on the dashboard
        """
        # Flatten the IMAQ image to a JPEG
        #t0 = time.time()
        vdata, size = nivision.imaqFlatten(image,
                nivision.IMAQ_FLATTEN_IMAGE,
                nivision.IMAQ_COMPRESSION_JPEG, 10 * self.quality)
        data = (ctypes.c_ubyte * size).from_address(vdata)
        #print("flatten: %s" % (time.time() - t0))

        # If "HW" client setting, Find the start of the JPEG data
        #print(size, [x for x in data])
        start = 0
        with self.mutex:
            hwClient = True#self.hwClient
        if hwClient:
            while start < size - 1:
                if data[start] == 0xFF and data[start + 1] == 0xD8:
                    break
                start += 1
            #print(size, start)

        size -= start
        if size <= 2:
            raise ValueError("data size of flattened image is less than 2. Try another camera!")

        self._setImageData(data, size, start, vdata)

    def startAutomaticCapture(self, camera):
        """Start automatically capturing images to send to the dashboard.

        You should call this method to just see a camera feed on the dashboard
        without doing any vision processing on the roboRIO. {@link #setImage}
        shouldn't be called after this is called.

        :param camera: The camera interface (e.g. a USBCamera instance)
        """
        if self.captureThread is not None:
            return

        self.camera = camera
        self.camera.startCapture()

        self.captureThread = threading.Thread(target=self._autoCapture, name="CaptureThread")
        self.captureThread.daemon = True
        self.captureThread.start()

    def _autoCapture(self):
        frame = nivision.imaqCreateImage(nivision.IMAQ_IMAGE_RGB, 0)

        while True:
            with self.mutex:
                hwClient = True#self.hwClient
                if hwClient:
                    data = self.dataPool[-1]
                    self.dataPool.pop()
            try:
                if hwClient:
                    size = self.camera.getImageData(data, self.kMaxImageSize)
                    self._setImageData(data, size)
                else:
                    self.camera.getImage(frame)
                    self.setImage(frame)
            except (ValueError, IndexError):
                logger.exception("getting image")
                if hwClient and not self.ready.is_set():
                    self.dataPool.append(data)
                time.sleep(0.1)

    def isAutoCaptureStarted(self):
        """check if auto capture is started
        """
        with self.mutex:
            return self.captureThread is not None

    def setSize(self, size):
        if size == self.kSize160x120:
            self.camera.setSize(160, 120)
        elif size == self.kSize320x240:
            self.camera.setSize(320, 240)
        elif size == self.kSize640x480:
            self.camera.setSize(640, 480)
        else:
            return

    def setQuality(self, quality):
        """Set the quality of the compressed image sent to the dashboard

        :param quality: The quality of the JPEG image, from 0 to 100
        """
        with self.mutex:
            if quality > 100:
                self.quality = 100
            elif quality < 0:
                self.quality = 0
            else:
                self.quality = quality

    def getQuality(self):
        """Get the quality of the compressed image sent to the dashboard

        :returns: The quality, from 0 to 100
        """
        with self.mutex:
            return self.quality

    def _serve(self):
        """Run the M-JPEG server.

        This function listens for a connection from the dashboard in a
        background thread, then sends back the M-JPEG stream.
        """

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(('', self.kPort))
        sock.listen(1)

        while True:
            try:
                conn, addr = sock.accept()

                s = conn.makefile('rwb')

                fps = self.intStruct.unpack(s.read(4))[0]
                compression = self.intStruct.unpack(s.read(4))[0]
                size = self.intStruct.unpack(s.read(4))[0]
                logger.info("Client connected: %d fps, %d compression, %d size"
                            % (fps, compression, size))

                with self.mutex:
                    doWait = self.camera is None

                if doWait:
                    logger.info("camera not ready yet, awaiting first image")
                    self.ready.wait()

                if compression == self.kHardwareCompression:
                    hwClient = True
                else:
                    hwClient = False
                    self.setQuality(100-compression)

                with self.mutex:
                    self.hwClient = hwClient
                    self.camera.setFPS(fps)
                self.setSize(size)

                period = 1.0 / fps
                while True:
                    t0 = time.time()

                    self.ready.wait()
                    self.ready.clear()

                    imageData = (None, 0, 0, None)
                    try:
                        with self.mutex:
                            imageData = self.imageData
                            self.imageData = (None, 0, 0, None)

                        if imageData[0] is None:
                            continue

                        size = imageData[1]
                        start = imageData[2]
                        data = (ctypes.c_ubyte * size).from_address(ctypes.addressof(imageData[0])+start)

                        #print("sending image")
                        s.write(self.kMagicNumber)
                        # write size of image
                        s.write(self.intStruct.pack(size))
                        s.write(data)
                    except IOError:
                        # print error to driverstation
                        logger.exception("error sending image")
                        break
                    finally:
                        self._freeImageData(imageData)

                    s.flush()
                    dt = time.time() - t0
                    #print(dt, period)
                    if dt < period:
                        time.sleep(period - dt)
            except IOError:
                # print error to driverstation
                logger.exception("error in conn")
                continue




#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .subsystem import Subsystem

from ..pidcontroller import PIDController

__all__ = ["PIDSubsystem"]

class PIDSubsystem(Subsystem):
    """This class is designed to handle the case where there is a Subsystem
    which uses a single {@link PIDController} almost constantly (for instance,
    an elevator which attempts to stay at a constant height).

    It provides some convenience methods to run an internal PIDController.
    It also allows access to the internal PIDController in order to give total
    control to the programmer.
    """

    def __init__(self, p, i, d, period=None, f=0.0, name=None):
        """Instantiates a PIDSubsystem that will use the given p, i and d
        values.
        It will use the class name as its name unless otherwise specified.
        It will also space the time between PID loop calculations to be equal
        to the given period.
        
        :param p: the proportional value
        :param i: the integral value
        :param d: the derivative value
        :param period: the time (in seconds) between calculations (optional)
        :param f: the feed forward value
        :param name: the name (optional)
        """
        super().__init__(name)
        if period is None:
            period = PIDController.kDefaultPeriod
        self.controller = PIDController(p, i, d, f, self.returnPIDInput,
                                        self.usePIDOutput, period)

    def getPIDController(self):
        """Returns the PIDController used by this PIDSubsystem.
        Use this if you would like to fine tune the pid loop.

        Notice that calling :meth:`setSetpoint` on the controller
        will not result in the setpoint being trimmed to be in
        the range defined by :meth:`setSetpointRange`.

        :returns: the :class:`.PIDController` used by this PIDSubsystem
        """
        return self.controller

    def setSetpointRelative(self, deltaSetpoint):
        """Adds the given value to the setpoint.
        If :meth:`setRange` was used, then the bounds will still be honored by
        this method.
        
        :param deltaSetpoint: the change in the setpoint
        """
        self.setSetpoint(self.getSetpoint() + deltaSetpoint)

    def setSetpoint(self, setpoint):
        """Sets the setpoint to the given value.  If :meth:`setRange` was called,
        then the given setpoint will be trimmed to fit within the range.
        
        :param setpoint: the new setpoint
        """
        self.controller.setSetpoint(setpoint)

    def getSetpoint(self):
        """Returns the setpoint.
        
        :returns: the setpoint
        """
        return self.controller.getSetpoint()

    def getPosition(self):
        """Returns the current position
        
        :returns: the current position
        """
        return self.returnPIDInput()

    def setInputRange(self, minimumInput, maximumInput):
        """Sets the maximum and minimum values expected from the input.

        :param minimumInput: the minimum value expected from the input
        :param maximumInput: the maximum value expected from the output
        """
        self.controller.setInputRange(minimumInput, maximumInput)

    def setOutputRange(self, minimumOutput, maximumOutput):
        """Sets the maximum and minimum values to write.

        :param minimumOutput: the minimum value to write to the output
        :param maximumOutput: the maximum value to write to the output
        """
        self.controller.setOutputRange(minimumOutput, maximumOutput)

    def setAbsoluteTolerance(self, t):
        """Set the absolute error which is considered tolerable for use with
        OnTarget.
        
        :param t: The absolute tolerance (same range as the PIDInput values)
        """
        self.controller.setAbsoluteTolerance(t)

    def setPercentTolerance(self, p):
        """Set the percentage error which is considered tolerable for use with
        OnTarget.
        
        :param p: The percentage tolerance (value of 15.0 == 15 percent)
        """
        self.controller.setPercentTolerance(p)

    def onTarget(self):
        """Return True if the error is within the percentage of the total
        input range, determined by setAbsoluteTolerance or setPercentTolerance.
        This assumes that the maximum and minimum input were set using
        setInput.
        
        :returns: True if the error is less than the tolerance
        """
        return self.controller.onTarget()

    def returnPIDInput(self):
        """Returns the input for the pid loop.

        It returns the input for the pid loop, so if this command was based
        off of a gyro, then it should return the angle of the gyro

        All subclasses of PIDSubsystem must override this method.

        This method will be called in a different thread then the Scheduler
        thread.

        :returns: the value the pid loop should use as input
        """
        raise NotImplementedError

    def usePIDOutput(self, output):
        """Uses the value that the pid loop calculated.  The calculated value
        is the "output" parameter.
        This method is a good time to set motor values, maybe something along
        the lines of `driveline.tankDrive(output, -output)`.

        All subclasses of PIDSubsystem should override this method.

        This method will be called in a different thread then the Scheduler
        thread.

        :param output: the value the pid loop calculated
        """
        pass

    def enable(self):
        """Enables the internal :class:`.PIDController`
        """
        self.controller.enable()

    def disable(self):
        """Disables the internal :class:`.PIDController`
        """
        self.controller.disable()

    def getSmartDashboardType(self):
        return "PIDSubsystem"

    def initTable(self, table):
        self.controller.initTable(table)
        super().initTable(table)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .command import Command

from ..pidcontroller import PIDController

__all__ = ["PIDCommand"]

class PIDCommand(Command):
    """This class defines a Command which interacts heavily with a PID loop.

    It provides some convenience methods to run an internal PIDController.
    It will also start and stop said PIDController when the PIDCommand is
    first initialized and ended/interrupted.
    """

    def __init__(self, p, i, d, period=None, f=0.0, name=None):
        """Instantiates a PIDCommand that will use the given p, i and d values.
        It will use the class name as its name unless otherwise specified.
        It will also space the time between PID loop calculations to be equal
        to the given period.
        
        :param p: the proportional value
        :param i: the integral value
        :param d: the derivative value
        :param period: the time (in seconds) between calculations (optional)
        :param f: the feed forward value
        :param name: the name (optional)
        """
        super().__init__(name)
        self.controller = PIDController(p, i, d, f, self.returnPIDInput,
                                        self.usePIDOutput, period)

    def getPIDController(self):
        """Returns the PIDController used by this PIDCommand.
        Use this if you would like to fine tune the pid loop.

        Notice that calling setSetpoint(...) on the controller
        will not result in the setpoint being trimmed to be in
        the range defined by setSetpointRange(...).

        :returns: the PIDController used by this PIDCommand
        """
        return self.controller

    def _initialize(self):
        self.controller.enable()

    def _end(self):
        self.controller.disable()

    def _interrupted(self):
        self._end()

    def setSetpointRelative(self, deltaSetpoint):
        """Adds the given value to the setpoint.
        If :meth:`setRange` was used, then the bounds will still be honored by
        this method.
        
        :param deltaSetpoint: the change in the setpoint
        """
        self.setSetpoint(self.getSetpoint() + deltaSetpoint)

    def setSetpoint(self, setpoint):
        """Sets the setpoint to the given value.  If :meth:`setRange` was called,
        then the given setpoint will be trimmed to fit within the range.
        
        :param setpoint: the new setpoint
        """
        self.controller.setSetpoint(setpoint)

    def getSetpoint(self):
        """Returns the setpoint.
        
        :returns: the setpoint
        """
        return self.controller.getSetpoint()

    def getPosition(self):
        """Returns the current position
        
        :returns: the current position
        """
        return self.returnPIDInput()

    def returnPIDInput(self):
        """Returns the input for the pid loop.

        It returns the input for the pid loop, so if this command was based
        off of a gyro, then it should return the angle of the gyro

        All subclasses of PIDCommand must override this method.

        This method will be called in a different thread then the :class:`.Scheduler`
        thread.

        :returns: the value the pid loop should use as input
        """
        raise NotImplementedError

    def usePIDOutput(self, output):
        """Uses the value that the pid loop calculated.  The calculated value
        is the "output" parameter.
        This method is a good time to set motor values, maybe something along
        the lines of `driveline.tankDrive(output, -output)`.

        All subclasses of PIDCommand should override this method.

        This method will be called in a different thread then the Scheduler
        thread.

        :param output: the value the pid loop calculated
        """
        pass

    def getSmartDashboardType(self):
        return "PIDCommand"

    def initTable(self, table):
        self.controller.initTable(table)
        super().initTable(table)



# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.

import threading

import hal
import sys
import traceback

from .motorsafety import MotorSafety
from .timer import Timer

__all__ = ["DriverStation"]

JOYSTICK_UNPLUGGED_MESSAGE_INTERVAL = 1.0

class DriverStation:
    """Provide access to the network communication data to / from the Driver
    Station."""

    #: The number of joystick ports
    kJoystickPorts = 6

    class Alliance:
        """The robot alliance that the robot is a part of"""
        Red = 0
        Blue = 1
        Invalid = 2
        
    @staticmethod
    def _reset():
        if hasattr(DriverStation, 'instance'):
            ds = DriverStation.instance
            ds.release()
            hal.giveMultiWait(ds.packetDataAvailableSem)
            ds.thread.join()
            del DriverStation.instance

    @staticmethod
    def getInstance():
        """Gets the global instance of the DriverStation

        :returns: :class:`DriverStation`
        """
        if not hasattr(DriverStation, "instance"):
            DriverStation.instance = None
            DriverStation.instance = DriverStation()
        return DriverStation.instance

    def __init__(self):
        """DriverStation constructor.

        The single DriverStation instance is created statically with the
        instance static member variable, you should never create a
        DriverStation instance.
        """
        
        if not hasattr(DriverStation, 'instance') or DriverStation.instance is not None:
            raise ValueError("Do not create DriverStation instances, use DriverStation.getInstance() instead")
        
        self.mutex = threading.RLock()
        self.dataSem = threading.Condition(self.mutex)

        self.packetDataAvailableMutex = hal.initializeMutexNormal()
        self.packetDataAvailableSem = hal.initializeMultiWait()
        hal.HALSetNewDataSem(self.packetDataAvailableSem)

        self.nextMessageTime = 0.0

        self.joystickAxes = []
        self.joystickPOVs = []
        self.joystickButtons = []
        for i in range(self.kJoystickPorts):
            self.joystickAxes.append([0]*hal.kMaxJoystickAxes)
            self.joystickPOVs.append([0]*hal.kMaxJoystickPOVs)
            self.joystickButtons.append(hal.HALJoystickButtons())

        self.userInDisabled = False
        self.userInAutonomous = False
        self.userInTeleop = False
        self.userInTest = False
        self.newControlData = False

        self.thread_keepalive = True

        self.thread = threading.Thread(target=self.task, name="FRCDriverStation")
        self.thread.daemon = True
        self.thread.start()

    def __del__(self):
        hal.deleteMultiWait(self.packetDataAvailableSem)
        hal.deleteMutex(self.packetDataAvailableMutex)

    def release(self):
        """Kill the thread"""
        self.thread_keepalive = False

    def task(self):
        """Provides the service routine for the DS polling thread."""
        safetyCounter = 0
        while self.thread_keepalive:
            hal.takeMultiWait(self.packetDataAvailableSem,
                              self.packetDataAvailableMutex, 0)
            self.getData()
            with self.dataSem:
                self.dataSem.notify_all()
            safetyCounter += 1
            if safetyCounter >= 4:
                MotorSafety.checkMotors()
                safetyCounter = 0
            if self.userInDisabled:
                hal.HALNetworkCommunicationObserveUserProgramDisabled()
            if self.userInAutonomous:
                hal.HALNetworkCommunicationObserveUserProgramAutonomous()
            if self.userInTeleop:
                hal.HALNetworkCommunicationObserveUserProgramTeleop()
            if self.userInTest:
                hal.HALNetworkCommunicationObserveUserProgramTest()

    def waitForData(self, timeout = None):
        """Wait for new data or for timeout, which ever comes first.  If
        timeout is None, wait for new data only.

        :param timeout: The maximum time in milliseconds to wait.
        """
        with self.dataSem:
            self.dataSem.wait(timeout)

    def getData(self):
        """Copy data from the DS task for the user.
        If no new data exists, it will just be returned, otherwise
        the data will be copied from the DS polling loop.
        """
        with self.mutex:
            # Get the status of all of the joysticks
            for stick in range(self.kJoystickPorts):
                self.joystickAxes[stick] = hal.HALGetJoystickAxes(stick)
                self.joystickPOVs[stick] = hal.HALGetJoystickPOVs(stick)
                self.joystickButtons[stick] = hal.HALGetJoystickButtons(stick)
            self.newControlData = True

    def getBatteryVoltage(self):
        """Read the battery voltage.

        :returns: The battery voltage in Volts."""
        return hal.getVinVoltage()

    def _reportJoystickUnpluggedError(self, message):
        """
        Reports errors related to unplugged joysticks and throttles them so that they don't overwhelm the DS
        """
        currentTime = Timer.getFPGATimestamp()
        if currentTime > self.nextMessageTime:
            self.reportError(message, False)
            self.nextMessageTime = currentTime + JOYSTICK_UNPLUGGED_MESSAGE_INTERVAL

    def getStickAxis(self, stick, axis):
        """Get the value of the axis on a joystick.
        This depends on the mapping of the joystick connected to the specified
        port.

        :param stick: The joystick port number
        :param axis: The analog axis value to read from the joystick.
        :returns: The value of the axis on the joystick.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        if axis < 0 or axis >= hal.kMaxJoystickAxes:
            raise IndexError("Joystick axis is out of range")

        with self.mutex:
            joystickAxes = self.joystickAxes[stick]

            if axis >= len(joystickAxes):
                self._reportJoystickUnpluggedError("WARNING: Joystick axis %d on port %d not available, check if controller is plugged in\n" % (axis, stick))
                return 0.0
            value = joystickAxes[axis]
        if value < 0:
            return value / 128.0
        else:
            return value / 127.0

    def getStickAxisCount(self, stick):
        """Returns the number of axes on a given joystick port

        :param stick: The joystick port number

        :returns: The number of axes on the indicated joystick
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.mutex:
            return len(self.joystickAxes[stick])

    def getStickPOV(self, stick, pov):
        """Get the state of a POV on the joystick.

        :param stick: The joystick port number
        :param pov: which POV
        :returns: The angle of the POV in degrees, or -1 if the POV is not
                  pressed.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        if pov < 0 or pov >= hal.kMaxJoystickPOVs:
            raise IndexError("Joystick POV is out of range")

        with self.mutex:
            joystickPOVs = self.joystickPOVs[stick]

            if pov >= len(joystickPOVs):
                self._reportJoystickUnpluggedError("WARNING: Joystick POV %d on port %d not available, check if controller is plugged in\n" % (pov, stick))
                return 0.0
            return joystickPOVs[pov]

    def getStickPOVCount(self, stick):
        """Returns the number of POVs on a given joystick port

        :param stick: The joystick port number

        :returns: The number of POVs on the indicated joystick
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.mutex:
            return len(self.joystickPOVs[stick])

    def getStickButtons(self, stick):
        """The state of all the buttons on the joystick.

        :param stick: The joystick port number
        :returns: The state of all buttons, as a bit array.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.mutex:
            return self.joystickButtons[stick]

    def getStickButton(self, stick, button):
        """The state of a button on the joystick.

        :param stick: The joystick port number
        :param button: The button number to be read.
        :returns: The state of the button.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.mutex:
            buttons = self.joystickButtons[stick]
            if button > buttons.count:
                self._reportJoystickUnpluggedError("WARNING: Joystick Button %d on port %d not available, check if controller is plugged in\n" % (button, stick))
                return False
            if button <= 0:
                self._reportJoystickUnpluggedError("ERROR: Button indexes begin at 1 for WPILib\n")
                return False
            return ((0x1 << (button - 1)) & buttons.buttons) != 0

    def getStickButtonCount(self, stick):
        """Gets the number of buttons on a joystick

        :param stick: The joystick port number

        :returns: The number of buttons on the indicated joystick.
        """
        if stick < 0 or stick >= self.kJoystickPorts:
            raise IndexError("Joystick index is out of range, should be 0-%s" % self.kJoystickPorts)

        with self.mutex:
            return self.joystickButtons[stick].count

    def isEnabled(self):
        """Gets a value indicating whether the Driver Station requires the
        robot to be enabled.

        :returns: True if the robot is enabled, False otherwise.
        """
        controlWord = hal.HALGetControlWord()
        return controlWord.enabled != 0 and controlWord.dsAttached != 0

    def isDisabled(self):
        """Gets a value indicating whether the Driver Station requires the
        robot to be disabled.

        :returns: True if the robot should be disabled, False otherwise.
        """
        return not self.isEnabled()

    def isAutonomous(self):
        """Gets a value indicating whether the Driver Station requires the
        robot to be running in autonomous mode.

        :returns: True if autonomous mode should be enabled, False otherwise.
        """
        controlWord = hal.HALGetControlWord()
        return controlWord.autonomous != 0

    def isTest(self):
        """Gets a value indicating whether the Driver Station requires the
        robot to be running in test mode.

        :returns: True if test mode should be enabled, False otherwise.
        """
        controlWord = hal.HALGetControlWord()
        return controlWord.test != 0

    def isOperatorControl(self):
        """Gets a value indicating whether the Driver Station requires the
        robot to be running in operator-controlled mode.

        :returns: True if operator-controlled mode should be enabled,
            False otherwise.
        """
        controlWord = hal.HALGetControlWord()
        return not (controlWord.autonomous != 0 or controlWord.test != 0)

    def isSysActive(self):
        """
        Gets a value indicating whether the FPGA outputs are enabled. The outputs may be disabled
        if the robot is disabled or e-stopped, the watdhog has expired, or if the roboRIO browns out.

        :returns: True if the FPGA outputs are enabled.
        """
        return hal.HALGetSystemActive()

    def isBrownedOut(self):
        """
        Check if the system is browned out.

        :returns: True if the system is browned out.
        """
        return hal.HALGetBrownedOut()

    def isNewControlData(self):
        """Has a new control packet from the driver station arrived since the
        last time this function was called?

        :returns: True if the control data has been updated since the last
            call.
        """
        with self.mutex:
            result = self.newControlData
            self.newControlData = False
            return result

    def getAlliance(self):
        """Get the current alliance from the FMS.

        :returns: The current alliance
        :rtype: :class:`DriverStation.Alliance`
        """
        allianceStationID = hal.HALGetAllianceStation()
        if allianceStationID in (hal.kHALAllianceStationID_red1,
                                 hal.kHALAllianceStationID_red2,
                                 hal.kHALAllianceStationID_red3):
            return self.Alliance.Red
        elif allianceStationID in (hal.kHALAllianceStationID_blue1,
                                   hal.kHALAllianceStationID_blue2,
                                   hal.kHALAllianceStationID_blue3):
            return self.Alliance.Blue
        else:
            return self.Alliance.Invalid

    def getLocation(self):
        """Gets the location of the team's driver station controls.

        :returns: The location of the team's driver station controls:
            1, 2, or 3
        """
        allianceStationID = hal.HALGetAllianceStation()
        if allianceStationID in (hal.kHALAllianceStationID_red1,
                                 hal.kHALAllianceStationID_blue1):
            return 1
        elif allianceStationID in (hal.kHALAllianceStationID_red2,
                                   hal.kHALAllianceStationID_blue2):
            return 2
        elif allianceStationID in (hal.kHALAllianceStationID_red3,
                                   hal.kHALAllianceStationID_blue3):
            return 3
        else:
            return 0

    def isFMSAttached(self):
        """Is the driver station attached to a Field Management System?

        :returns: True if the robot is competing on a field being controlled
            by a Field Management System
        """
        controlWord = hal.HALGetControlWord()
        return controlWord.fmsAttached != 0

    def isDSAttached(self):
        """Is the driver station attached to the robot?

        :returns: True if the robot is being controlled by a driver station.
        """
        controlWord = hal.HALGetControlWord()
        return controlWord.dsAttached != 0

    def getMatchTime(self):
        """Return the approximate match time.
        The FMS does not currently send the official match time to the robots, but
        does send an approximate match time. The value will count down the time
        remaining in the current period (auto or teleop).

        .. warning::

            This is not an official time (so it cannot be used to argue with
            referees or guarantee that a function will trigger before a match ends).

        The Practice Match function of the DS approximates the behaviour seen on the field.

        :returns: Time remaining in current match period (auto or teleop) in seconds
        """
        return hal.HALGetMatchTime()

    @staticmethod
    def reportError(error, printTrace):
        """Report error to Driver Station, and also prints error to `sys.stderr`.
        Optionally appends stack trace to error message.

        :param printTrace: If True, append stack trace to error string
        """
        errorString = error
        if printTrace:
            exc = sys.exc_info()[0]
            stack = traceback.extract_stack()[:-1]  # last one is this func
            if exc is not None: # i.e. if an exception is present
                # remove call of full_stack, the printed exception
                # will contain the caught exception caller instead
                del stack[-1]
            trc = 'Traceback (most recent call last):\n'
            stackstr = trc + ''.join(traceback.format_list(stack))
            if exc is not None:
                stackstr += '  ' + traceback.format_exc().lstrip(trc)
            errorString += ':\n' + stackstr
        #print(errorString, file=sys.stderr)
        controlWord = hal.HALGetControlWord()
        if controlWord.dsAttached != 0:
            hal.HALSetErrorData(errorString, 0)

    def InDisabled(self, entering):
        """Only to be used to tell the Driver Station what code you claim to
        be executing for diagnostic purposes only.

        :param entering: If True, starting disabled code; if False, leaving
            disabled code
        """
        self.userInDisabled = entering

    def InAutonomous(self, entering):
        """Only to be used to tell the Driver Station what code you claim to
        be executing for diagnostic purposes only.

        :param entering: If True, starting autonomous code; if False, leaving
            autonomous code
        """
        self.userInAutonomous = entering

    def InOperatorControl(self, entering):
        """Only to be used to tell the Driver Station what code you claim to
        be executing for diagnostic purposes only.

        :param entering: If True, starting teleop code; if False, leaving
            teleop code
        """
        self.userInTeleop = entering

    def InTest(self, entering):
        """Only to be used to tell the Driver Station what code you claim to
        be executing for diagnostic purposes only.

        :param entering: If True, starting test code; if False, leaving test
            code
        """
        self.userInTest = entering



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2014. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref

from .livewindowsendable import LiveWindowSendable
from .resource import Resource
from .sensorbase import SensorBase

__all__ = ["PWM"]

def _freePWM(port):
    hal.setPWM(port, 0)
    hal.freePWMChannel(port)
    hal.freeDIO(port)

class PWM(LiveWindowSendable):
    """Raw interface to PWM generation in the FPGA.
    
    The values supplied as arguments for PWM outputs range from -1.0 to 1.0. They
    are mapped to the hardware dependent values, in this case 0-2000 for the
    FPGA.  Changes are immediately sent to the FPGA, and the update occurs at
    the next FPGA cycle. There is no delay.

    As of revision 0.1.10 of the FPGA, the FPGA interprets the 0-2000 values as
    follows:
    
    - 2000 = full "forward"
    - 1999 to 1001 = linear scaling from "full forward" to "center"
    - 1000 = center value
    - 999 to 2 = linear scaling from "center" to "full reverse"
    - 1 = minimum pulse width (currently .5ms)
    - 0 = disabled (i.e. PWM output is held low)
    
    kDefaultPwmPeriod is the 1x period (5.05 ms).  In hardware, the period
    scaling is implemented as an output squelch to get longer periods for old
    devices.

    - 20ms periods (50 Hz) are the "safest" setting in that this works for all
      devices
    - 20ms periods seem to be desirable for Vex Motors
    - 20ms periods are the specified period for HS-322HD servos, but work
      reliably down to 10.0 ms; starting at about 8.5ms, the servo sometimes
      hums and get hot; by 5.0ms the hum is nearly continuous
    - 10ms periods work well for Victor 884
    - 5ms periods allows higher update rates for Luminary Micro Jaguar speed
      controllers. Due to the shipping firmware on the Jaguar, we can't run the
      update period less than 5.05 ms.
      
    .. not_implemented: initPWM
    """
    class PeriodMultiplier:
        """Represents the amount to multiply the minimum servo-pulse pwm
        period by.
        """
        k1X = 1
        k2X = 2
        k4X = 4

    #: the default PWM period measured in ms.
    kDefaultPwmPeriod = 5.05
    
    #: the PWM range center in ms
    kDefaultPwmCenter = 1.5
    
    #: the number of PWM steps below the centerpoint
    kDefaultPwmStepsDown = 1000
    
    #: the value to use to disable
    kPwmDisabled = 0

    def __init__(self, channel):
        """Allocate a PWM given a channel.

        :param channel: The PWM channel number. 0-9 are on-board, 10-19 are on the MXP port
        :type channel: int
        """
        SensorBase.checkPWMChannel(channel)
        self.channel = channel
        self._port = hal.initializeDigitalPort(hal.getPort(channel))

        if not hal.allocatePWMChannel(self._port):
            raise IndexError("PWM channel %d is already allocated" % channel)
        
        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

        hal.setPWM(self._port, 0)

        self._pwm_finalizer = weakref.finalize(self, _freePWM, self._port)

        self.eliminateDeadband = False

        hal.HALReport(hal.HALUsageReporting.kResourceType_PWM, channel)

    @property
    def port(self):
        if not self._pwm_finalizer.alive:
            return None
        return self._port

    def free(self):
        """Free the PWM channel.

        Free the resource associated with the PWM channel and set the value
        to 0.
        """
        self._pwm_finalizer()

    def enableDeadbandElimination(self, eliminateDeadband):
        """Optionally eliminate the deadband from a speed controller.

        :param eliminateDeadband: If True, set the motor curve on the Jaguar
            to eliminate the deadband in the middle of the range. Otherwise, keep
            the full range without modifying any values.
        :type eliminateDeadband: bool
        """
        self.eliminateDeadband = eliminateDeadband

    def setBounds(self, max, deadbandMax, center, deadbandMin, min):
        """Set the bounds on the PWM pulse widths.

        This sets the bounds on the PWM values for a particular type of
        controller.  The values determine the upper and lower speeds as well
        as the deadband bracket.

        :param max: The max PWM pulse width in ms
        :type max: float
        :param deadbandMax: The high end of the deadband range pulse width in ms
        :type deadbandMax: float
        :param center: The center (off) pulse width in ms
        :type center: float
        :param deadbandMin: The low end of the deadband pulse width in ms
        :type deadbandMin: float
        :param min: The minimum pulse width in ms
        :type min: float
        """
        loopTime = hal.getLoopTiming()/(SensorBase.kSystemClockTicksPerMicrosecond*1e3)

        self.maxPwm = int((max-self.kDefaultPwmCenter)/loopTime+self.kDefaultPwmStepsDown-1)
        self.deadbandMaxPwm = int((deadbandMax-self.kDefaultPwmCenter)/loopTime+self.kDefaultPwmStepsDown-1)
        self.centerPwm = int((center-self.kDefaultPwmCenter)/loopTime+self.kDefaultPwmStepsDown-1)
        self.deadbandMinPwm = int((deadbandMin-self.kDefaultPwmCenter)/loopTime+self.kDefaultPwmStepsDown-1)
        self.minPwm = int((min-self.kDefaultPwmCenter)/loopTime+self.kDefaultPwmStepsDown-1)

    def getChannel(self):
        """Gets the channel number associated with the PWM Object.

        :returns: The channel number.
        :rtype: int
        """
        return self.channel

    def setPosition(self, pos):
        """Set the PWM value based on a position.

        This is intended to be used by servos.

        .. note::

            :func:`setBounds` must be called first.

        :param pos: The position to set the servo between 0.0 and 1.0.
        :type pos: float
        """
        if pos < 0.0:
            pos = 0.0
        elif pos > 1.0:
            pos = 1.0

        rawValue = int(pos * self.getFullRangeScaleFactor() + self.getMinNegativePwm())

        # send the computed pwm value to the FPGA
        self.setRaw(rawValue)

    def getPosition(self):
        """Get the PWM value in terms of a position.

        This is intended to be used by servos.

        .. note::

            :func:`setBounds` must be called first.

        :returns: The position the servo is set to between 0.0 and 1.0.
        :rtype: float
        """
        value = self.getRaw()
        if value < self.getMinNegativePwm():
            return 0.0
        elif value > self.getMaxPositivePwm():
            return 1.0
        else:
            return float(value - self.getMinNegativePwm()) / self.getFullRangeScaleFactor()

    def setSpeed(self, speed):
        """Set the PWM value based on a speed.

        This is intended to be used by speed controllers.

        .. note::

            :func:`setBounds` must be called first.

        :param speed: The speed to set the speed controller between -1.0 and
            1.0.
        :type speed: float
        """
        # clamp speed to be in the range 1.0 >= speed >= -1.0
        if speed < -1.0:
            speed = -1.0
        elif speed > 1.0:
            speed = 1.0

        # calculate the desired output pwm value by scaling the speed
        # appropriately
        if speed == 0.0:
            rawValue = self.getCenterPwm()
        elif speed > 0.0:
            rawValue = int(speed * self.getPositiveScaleFactor() +
                           self.getMinPositivePwm() + 0.5)
        else:
            rawValue = int(speed * self.getNegativeScaleFactor() +
                           self.getMaxNegativePwm() + 0.5)

        # send the computed pwm value to the FPGA
        self.setRaw(rawValue)

    def getSpeed(self):
        """Get the PWM value in terms of speed.

        This is intended to be used by speed controllers.

        .. note::

            :func:`setBounds` must be called first.

        :returns: The most recently set speed between -1.0 and 1.0.
        :rtype: float
        """
        value = self.getRaw()
        if value > self.getMaxPositivePwm():
            return 1.0
        elif value < self.getMinNegativePwm():
            return -1.0
        elif value > self.getMinPositivePwm():
            return float(value - self.getMinPositivePwm()) / self.getPositiveScaleFactor()
        elif value < self.getMaxNegativePwm():
            return float(value - self.getMaxNegativePwm()) / self.getNegativeScaleFactor()
        else:
            return 0.0

    def setRaw(self, value):
        """Set the PWM value directly to the hardware.

        Write a raw value to a PWM channel.

        :param value: Raw PWM value.  Range 0 - 255.
        :type value: int
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        hal.setPWM(self.port, value)

    def getRaw(self):
        """Get the PWM value directly from the hardware.

        Read a raw value from a PWM channel.

        :returns: Raw PWM control value.  Range: 0 - 255.
        :rtype: int
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        return hal.getPWM(self.port)

    def setPeriodMultiplier(self, mult):
        """Slow down the PWM signal for old devices.

        :param mult: The period multiplier to apply to this channel
        :type mult: PWM.PeriodMultiplier
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        if mult == PWM.PeriodMultiplier.k4X:
            # Squelch 3 out of 4 outputs
            hal.setPWMPeriodScale(self.port, 3)
        elif mult == PWM.PeriodMultiplier.k2X:
            # Squelch 1 out of 2 outputs
            hal.setPWMPeriodScale(self.port, 1)
        elif mult == PWM.PeriodMultiplier.k1X:
            # Don't squelch any outputs
            hal.setPWMPeriodScale(self.port, 0)
        else:
            raise ValueError("Invalid mult argument '%s'" % mult)

    def setZeroLatch(self):
        if self.port is None:
            raise ValueError("operation on freed port")
        hal.latchPWMZero(self.port)

    def getMaxPositivePwm(self):
        return self.maxPwm

    def getMinPositivePwm(self):
        if self.eliminateDeadband:
            return self.deadbandMaxPwm
        else:
            return self.centerPwm + 1

    def getCenterPwm(self):
        return self.centerPwm

    def getMaxNegativePwm(self):
        if self.eliminateDeadband:
            return self.deadbandMinPwm
        else:
            return self.centerPwm - 1

    def getMinNegativePwm(self):
        return self.minPwm

    def getPositiveScaleFactor(self):
        """Get the scale for positive speeds."""
        return self.getMaxPositivePwm() - self.getMinPositivePwm()

    def getNegativeScaleFactor(self):
        """Get the scale for negative speeds."""
        return self.getMaxNegativePwm() - self.getMinNegativePwm()

    def getFullRangeScaleFactor(self):
        """Get the scale for positions."""
        return self.getMaxPositivePwm() - self.getMinNegativePwm()

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Speed Controller"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.getSpeed())

    def valueChanged(self, itable, key, value, bln):
        self.setSpeed(float(value))

    def startLiveWindowMode(self):
        self.setSpeed(0) # Stop for safety
        super().startLiveWindowMode()

    def stopLiveWindowMode(self):
        super().stopLiveWindowMode()
        self.setSpeed(0) # Stop for safety
